import sys
from importlib import import_module
from abc import ABC, abstractmethod
from datetime import datetime

import gzip
import bz2

import json

from .base_cov import Covariance
from .util import Log, deserialize, ensure_2d, make_multi_time_argument
from .derivatives import (
    gradient,
    hessian,
    hessian_log_determinant,
)
from .validation import _validate_time_x, _validate_float, _validate_array


logger = Log()


class Predictor(ABC):
    """
    Abstract base class for predictor models. It provides a common interface for all subclasses, which are expected to
    implement the `__call__` method for making predictions.

    An instance `predictor` of a subclass of `Predictor` can be used to make a prediction by calling it with input data `x`:

    >>> y = predictor(x)

    It is the responsibility of subclasses to define the behaviour of `_predict`.

    Methods
    -------
    __call__(x: Union[array-like, pd.DataFrame]):
        This makes predictions for an input `x`. The input data type can be either an array-like object
        (like list or numpy array) or a pandas DataFrame.

        Parameters
        ----------
        x : array-like or pandas.DataFrame
            The input data on which to make a prediction.

        Returns
        -------
        This method returns the predictions made by the model on the input data `x`.
        The prediction is usually array-like with as many entries as `x.shape[0]`.
    """

    # number of features of input data (x.shape[1]) to be specified in __init__
    n_input_features: int

    @abstractmethod
    def __init__(self):
        """Initialize the predictor. Must be overridden by subclasses."""
        pass

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        string = (
            'A predictor of class "'
            + self.__class__.__name__
            + '" with covariance function "'
            + repr(self.cov_func)
            + '" and data:\n'
            + "\n".join(
                [
                    str(key) + ": " + repr(getattr(self, key))
                    for key in self._data_dict().keys()
                ]
            )
        )
        return string

    @abstractmethod
    def _predict(self, *args, **kwars):
        """Call the predictor. Must be overridden by subclasses."""

    def __call__(self, x):
        """
        Use the trained model to make a prediction based on the input array, x.

        The input array should be 2D with its second dimension's length
        equal to the number of features used in training the model.

        Parameters
        ----------
        x : array-like
            The input data to the predictor.
            The array should have shape (n_samples, n_input_features).

        Returns
        -------
        array
            The predicted output generated by the model.

        Raises
        ------
        ValueError
            If the number of features in 'x' does not match the
            number of features the predictor was trained on.
        """
        x = _validate_array(x, "x")
        x = ensure_2d(x)
        if x.shape[1] != self.n_input_features:
            raise ValueError(
                f"The predictor was trained on data with {self.n_input_features} features. "
                f"However, the provided input data has {x.shape[1]} features. "
                "Please ensure that the input data has the same number of features as the training data."
            )
        return self._predict(x)

    @abstractmethod
    def _data_dict(self):
        """Return a dictionary containing the predictor's state data.
        All arrays nee to be numpy arrays for serialization.

        This method must be implemented by subclasses. It should return a
        dictionary where each key-value pair corresponds to an attribute of
        the predictor and its current state.

        :return: A dictionary containing the predictor's state data.
        :rtype: dict
        """
        pass

    def gradient(self, x, jit=True):
        R"""
        Conputes the gradient of the predict function for each line in x.

        :param x: Data points.
        :type x: array-like
        :param jit: Use jax just in time compilation. Defaults to True.
        :type jit: bool
        :return: gradiants - The gradient of function at each point in x.
            gradients.shape == x.shape
        :rtype: array-like
        """
        x = _validate_array(x, "x")
        x = ensure_2d(x)
        return gradient(self.__call__, x, jit=jit)

    def hessian(self, x, jit=True):
        R"""
        Conputes the hessian of the predict function for each line in x.

        :param x: Data points.
        :type x: array-like
        :param jit: Use jax just in time compilation. Defaults to True.
        :type jit: bool
        :return: hessians - The hessian matrix of function at each point in x.
            hessians.shape == X.shape + X.shape[1:]
        :rtype: array-like
        """
        x = _validate_array(x, "x")
        x = ensure_2d(x)
        return hessian(self.__call__, x, jit=jit)

    def hessian_log_determinant(self, x, jit=True):
        R"""
        Conputes the logarirhm of the determinat of the predict function for
        each line in x.

        :param x: Data points.
        :type x: array-like
        :param jit: Use jax just in time compilation. Defaults to True.
        :type jit: bool
        :return: signs, log_determinants - The sign of the determinant
            at each point x and the logarithm of its absolute value.
            signs.shape == log_determinants.shape == x.shape[0]
        :rtype: array-like, array-like
        """
        x = _validate_array(x, "x")
        x = ensure_2d(x)
        return hessian_log_determinant(self.__call__, x, jit=jit)

    def __getstate__(self):
        """Get the current state of the predictor for serialization.

        This method is called when serializing the predictor instance. It
        returns a dictionary that represents the current state of the
        predictor.

        :return: A dictionary representing the state of the predictor.
        :rtype: dict
        """
        module_name = self.__class__.__module__
        module = import_module(module_name)
        version = getattr(module, "__version__", "NA")
        data = self._data_dict()
        data.update({"n_input_features": self.n_input_features})

        state = {
            "data": data,
            "cov_func": self.cov_func.__getstate__(),
            "metadata": {
                "classname": self.__class__.__name__,
                "module_name": module_name,
                "module_version": version,
                "serialization_date": datetime.now().isoformat(),
                "python_version": sys.version,
            },
        }
        return state

    def __setstate__(self, state):
        """Set the current state of the predictor from deserialized state.

        This method is called when deserializing the predictor instance. It
        takes a dictionary that represents the state of the predictor and
        updates the predictor's attributes accordingly.

        :param state: A dictionary representing the state of the predictor.
        :type state: dict
        """
        data = state["data"]
        for name, value in data.items():
            val = deserialize(value)
            setattr(self, name, val)
        self.cov_func = Covariance.from_dict(state["cov_func"])

    def to_json(self, filename=None, compress=None):
        """Serialize the predictor to a JSON file.

        This method serializes the predictor to a JSON file. It can optionally
        compress the JSON file using gzip or bz2 compression.
        It automatically detects the compression method based on the file extension
        or use the compress keyword to determine the compression method.
        It also makes sure the file is saved with the appropriate file extension.

        :param filename: The name of the JSON file to which to serialize the predictor.
            If filname is None then the JSON string is returned instead.
        :type filename: str or None
        :param compress: The compression method to use ('gzip' or 'bz2'). If None, no compression is used.
        :type compress: str, optional
        """
        json_str = json.dumps(self.to_dict())

        if filename is None:
            return json_str

        if compress == "gzip":
            if isinstance(filename, str) and not filename.endswith(".gz"):
                filename += ".gz"
            with gzip.open(filename, "wt") as f:
                f.write(json_str)
        elif compress == "bz2":
            if isinstance(filename, str) and not filename.endswith(".bz2"):
                filename += ".bz2"
            with bz2.open(filename, "wt") as f:
                f.write(json_str)
        elif compress is None:
            with open(filename, "w") as f:
                f.write(json_str)
        else:
            msg = (
                f"Unknown compression format {compress}.\n"
                'Availabe formats are "gzip", "bz2" and None.'
            )
            logger.error(msg)
            raise ValueError(msg)
        logger.info(f"Written predictor to {filename}.")

    def to_dict(self):
        """Serialize the predictor to a python dictionary.

        :return: A python dictionary with the predictor data.
        :rtype: dict
        """
        return self.__getstate__()

    @classmethod
    def from_json(cls, filepath, compress=None):
        """Deserialize the predictor from a JSON file.

        This method deserializes the predictor from a JSON file. It automatically
        detects the compression method based on the file extension or uses
        the compress keyword to determine the compression method.

        :param filename: The path of the JSON file from which to deserialize the predictor.
        :type filename: str, pathlib.Path, or os.path
        :param compress: The compression method to use ('gzip' or 'bz2'). If None, no compression is used.
        :type compress: str, optional
        :return: An instance of the predictor.
        :rtype: Predictor subclass instance
        """
        filename = str(filepath)
        if compress is None:
            compress = "none"
        if compress == "gzip" or filename.endswith(".gz"):
            open_func = gzip.open
        elif compress == "bz2" or filename.endswith(".bz2"):
            open_func = bz2.open
        else:
            open_func = open

        with open_func(filepath, "rt") as f:
            json_str = f.read()

        return cls.from_json_str(json_str)

    @classmethod
    def from_dict(cls, data_dict):
        """Deserialize the predictor from a python dictionay.

        This method deserializes the predictor from a python dictionary.

        :param data_dict: The dictionary from which to deserialize the predictor.
        :type data_dict: dict
        :return: An instance of the predictor.
        :rtype: Predictor subclass instance
        """
        return cls.from_json_str(json.dumps(data_dict))

    @classmethod
    def from_json_str(cls, json_str):
        """Deserialize the predictor from a JSON string.

        This method deserializes the predictor from the content of a JSON file.

        :param json_str: The JSON string from which to deserialize the predictor.
        :type json_str: str
        :return: An instance of the predictor.
        :rtype: Predictor subclass instance
        """
        state = json.loads(json_str)
        clsname = state["metadata"]["classname"]
        module_name = state["metadata"]["module_name"]

        module = import_module(module_name)
        Subclass = getattr(module, clsname)
        instance = Subclass.__new__(Subclass)
        instance.__setstate__(state)

        return instance


class PredictorTime(Predictor):
    """
    Abstract base class for predictor models with a time covariate.

    An instance `predictor` of a subclass of `PredictorTime` can be used to
    make a prediction by calling it with input data `x` and `time`:

    >>> y = predictor(x, time)

    It is the responsibility of subclasses to define the behaviour of `_predict`.

    Methods
    -------
    __call__(x: Union[array-like, pd.DataFrame]):
        This makes predictions for an input `x`. The input data type can be either an array-like object
        (like list or numpy array) or a pandas DataFrame.

        Parameters
        ----------
        x : array-like or pandas.DataFrame
            The input data on which to make a prediction.
        time : scalar or array-like, optional
            The time points associated with each cell/row in 'x'.
            If 'time' is a scalar, it will be converted into a 1D array of the same size as 'x'.

        Returns
        -------
        This method returns the predictions made by the model on the input data `x` and `time`.
        The prediction is usually array-like with as many entries as `x.shape[0]`.
    """

    @make_multi_time_argument
    def __call__(self, Xnew, time=None):
        """
        Call method to use the class instance as a function. This method
        deals with an optional 'time' argument.
        If 'time' is a scalar, it converts it to a 1D array of the same size as 'Xnew'.

        Parameters
        ----------
        Xnew : array-like
            The new data points for prediction.
        time : scalar or array-like, optional
            The time points associated with each cell/row in 'Xnew'.
            If 'time' is a scalar, it will be converted into a 1D array of the same size as 'Xnew'.
        multi_time : array-like, optional
            If 'multi_time' is specified then a prediction will be made for each row.

        Returns
        -------
        array-like
            Predictions for 'Xnew'.

        Raises
        ------
        ValueError
            If 'time' is an array and its size does not match 'Xnew'.
        """
        # if time is a scalar, convert it into a 1D array of the same size as Xnew
        Xnew = _validate_time_x(
            Xnew, time, n_features=self.n_input_features, cast_scalar=True
        )

        return self._predict(Xnew)

    @make_multi_time_argument
    def time_derivative(
        self,
        x,
        time,
        jit=True,
    ):
        R"""
        Computes the time derivative of the prediction function for each line in `x`.

        This function applies a jax-based gradient operation to the density
        function evaluated at a specific time.
        The derivative is with respect to time and not the inputs in `x`.

        Parameters
        ----------
        x : array-like
            Data points where the derivative is to be evaluated.
        time : array-like or scalar
            Time point or points at which to evaluate the derivative.
            If `time` is a scalar, the derivative will be computed at this
            specific time point for all data points in `x`.
            If `time` is an array, it should be 1-D and the time derivative
            will be computed for all data-points at the corresponding time in the array.
        multi_time : array-like, optional
            If 'multi_time' is specified then a prediction will be made for each row.
        jit : bool, optional
            If True, use JAX's just-in-time (JIT) compilation to speed up the computation. Defaults to True.

        Returns
        -------
        array-like
            The time derivative of the prediction function evaluated at each point in `x`.
            The shape of the output array is the same as `x`.

        """
        Xnew = _validate_time_x(
            x, time, n_features=self.n_input_features, cast_scalar=True
        )
        return super().gradient(Xnew)[:, -1]

    @make_multi_time_argument
    def gradient(self, x, time, jit=True):
        """
        Computes the gradient of the prediction function for each point in `x` at a given time.

        Parameters
        ----------
        x : array-like
            Data points at which the gradient is to be computed.
        time : float
            Specific time point at which to compute the gradient.
        multi_time : array-like, optional
            If 'multi_time' is specified then the computation will be made for each row.
        jit : bool, optional
            If True, use JAX's just-in-time (JIT) compilation to speed up the computation. Defaults to True.

        Returns
        -------
        array-like
            The gradient of the prediction function at each point in `x`.
            The shape of the output array is the same as `x`.
        """
        time = _validate_float(time, "time", optional=True)

        def dens_at(x):
            return self.__call__(x, time)

        return gradient(dens_at, x, jit=jit)

    @make_multi_time_argument
    def hessian(self, x, time, jit=True):
        """
        Computes the Hessian matrix of the prediction function for each point in `x` at a given time.

        Parameters
        ----------
        x : array-like
            Data points at which the Hessian matrix is to be computed.
        time : float
            Specific time point at which to compute the Hessian matrix.
        multi_time : array-like, optional
            If 'multi_time' is specified then the computation will be made for each row.
        jit : bool, optional
            If True, use JAX's just-in-time (JIT) compilation to speed up the computation. Defaults to True.

        Returns
        -------
        array-like
            The Hessian matrix of the prediction function at each point in `x`.
            The shape of the output array is `x.shape + x.shape[1:]`.
        """
        time = _validate_float(time, "time", optional=True)

        def dens_at(x):
            return self.__call__(x, time)

        return hessian(dens_at, x, jit=jit)

    @make_multi_time_argument
    def hessian_log_determinant(self, x, time, jit=True):
        """
        Computes the logarithm of the determinant of the Hessian of the prediction function
        for each point in `x` at a given time.

        Parameters
        ----------
        x : array-like
            Data points at which the log determinant is to be computed.
        time : float
            Specific time point at which to compute the log determinant.
                multi_time : array-like, optional
                        If 'multi_time' is specified then the computation will be made for each row.
        jit : bool, optional
            If True, use JAX's just-in-time (JIT) compilation to speed up the computation. Defaults to True.

        Returns
        -------
        array-like
            The sign of the determinant at each point in `x` and the logarithm of its absolute value.
            `signs.shape == log_determinants.shape == x.shape[0]`.
        """
        time = _validate_float(time, "time", optional=True)

        def dens_at(x):
            return self.__call__(x, time)

        return hessian_log_determinant(dens_at, x, jit=jit)



from jax.config import config as jaxconfig

from .model import (
    DensityEstimator,
    FunctionEstimator,
    DimensionalityEstimator,
    TimeSensitiveDensityEstimator,
)
from .base_predictor import Predictor
from .cov import Covariance
from .util import Log

from . import _util as util
from . import _cov as cov
from . import _parameters as parameters
from . import _inference as inference
from . import _conditional as conditional
from . import _derivatives as derivatives

__version__ = "1.3.0"

__all__ = [
    "DensityEstimator",
    "FunctionEstimator",
    "DimensionalityEstimator",
    "TimeSensitiveDensityEstimator",
    "Predictor",
    "Covariance",
    "Log",
    "util",
    "cov",
    "model",
    "parameters",
    "inference",
    "conditional",
    "decomposition",
    "derivatives",
    "__version__",
]

# Set default configuration at import time
jaxconfig.update("jax_enable_x64", True)
jaxconfig.update("jax_platform_name", "cpu")



import sys
from abc import ABC, abstractmethod
from importlib import import_module
from datetime import datetime
import json

from .util import Log, make_serializable, deserialize

MELLON_NAME = __name__.split(".")[0]

logger = Log()


class Covariance(ABC):
    R"""
    Base covariance function.
    """

    @abstractmethod
    def __init__(self):
        pass

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        """Return a string representation"""
        clsname = self.__class__.__name__
        arguments = [
            f"{key}={val}"
            for key, val in self.__dict__.items()
            if key != "active_dims" or val is not None
        ]
        return clsname + "(" + ", ".join(arguments) + ")"

    @abstractmethod
    def k(x, y):
        pass

    def __call__(self, x, y):
        return self.k(x, y)

    def __add__(self, other):
        return Add(self, other)

    def __radd__(self, other):
        return Add(self, other)

    def __mul__(self, other):
        return Mul(self, other)

    def __rmul__(self, other):
        return Mul(self, other)

    def __pow__(self, other):
        return Pow(self, other)

    def _data_dict(self):
        state = {key: make_serializable(val) for key, val in self.__dict__.items()}
        return state

    def __getstate__(self):
        """Get the current state of the covariance function for serialization.

        This method is called when serializing the predictor instance. It
        returns a dictionary that represents the current state of the
        predictor.

        :return: A dictionary representing the state of the predictor.
        :rtype: dict
        """
        module_name = self.__class__.__module__
        clsname = self.__class__.__name__
        if module_name == "__main__":
            logger.warning(
                f'The covariance function "{clsname}" is not part of {MELLON_NAME} '
                f"and seems to be user defined. Make sure the implementation "
                "is available for deserialization."
            )
        elif module_name.split(".")[0] != MELLON_NAME:
            logger.warning(
                f'The covariance function "{clsname}" is not part of {MELLON_NAME} '
                f'but of the module "{module_name}". Make sure the module '
                "is available for deserialization."
            )
        module = import_module(module_name)
        version = getattr(module, "__version__", "NA")
        state = {
            "type": "mellon.Covariance",
            "data": self._data_dict(),
            "metadata": {
                "classname": clsname,
                "module_name": module_name,
                "module_version": version,
                "serialization_date": datetime.now().isoformat(),
                "python_version": sys.version,
            },
        }
        return state

    def __setstate__(self, state):
        """Set the current state of the covariance from deserialized state.

        This method is called when deserializing the predictor instance. It
        takes a dictionary that represents the state of the predictor and
        updates the predictor's attributes accordingly.

        :param state: A dictionary representing the state of the predictor.
        :type state: dict
        """
        data = state["data"]
        for name, value in data.items():
            val = deserialize(value)
            setattr(self, name, val)

    def to_json(self):
        """Serialize the predictor to a JSON string.

        This method serializes the covariance function to a JSON string.

        :return: An JASON string.
        :rtype: string
        """
        return json.dumps(self.__getstate__())

    def to_dict(self):
        """Serialize the predictor to a python dictionary."""
        return self.__getstate__()

    @classmethod
    def from_json(cls, json_str):
        """Deserialize the covariance function from a JSON string.

        This method deserializes the predictor from a JSON file. It automatically
        detects the compression method based on the file extension.

        :param json_str: The JSON string from which to deserialize the covariance function.
        :type json_str: str
        :return: An instance of the covariance function.
        :rtype: Covariance subclass instance
        """

        state = json.loads(json_str)
        return cls.from_dict(state)

    @classmethod
    def from_dict(cls, state):
        """Deserialize the covariance function from a python dictionary.

        :param state: The python dictionary from which to deserialize the covariance function.
        :type state: dict
        :return: An instance of the covariance function.
        :rtype: Covariance subclass instance
        """
        if not isinstance(state, dict) or state.get("type") != "mellon.Covariance":
            raise ValueError(
                "The passed dict does not seem to define a covariance kernel."
            )
        clsname = state["metadata"]["classname"]
        module_name = state["metadata"]["module_name"]

        if clsname in globals():
            Subclass = globals()[clsname]
        else:
            module = import_module(module_name)
            Subclass = getattr(module, clsname)
        instance = Subclass.__new__(Subclass)
        instance.__setstate__(state)

        return instance


class CovariancePair(Covariance):
    R"""
    Supports combining two covariance functions with.
    """

    def __init__(self, left, right):
        self.left = left
        self.right = right

    @classmethod
    def k(self, x, y):
        """Implement the kernel that combines the two covariance functions
        self.left and self.right."""
        pass

    def __getstate__(self):
        """Get the current state of the covariance function for serialization.

        This method is called when serializing the covariance function instance. It
        returns a dictionary that represents the current state of the
        covariance function.

        :return: A dictionary representing the state of the covariance function.
        :rtype: dict
        """
        module_name = self.__class__.__module__.split(".")[0]
        module = import_module(module_name)
        version = getattr(module, "__version__", "NA")
        if callable(self.right):
            right_data = self.right.__getstate__()
        else:
            right_data = make_serializable(self.right)
        state = {
            "type": "mellon.Covariance",
            "left_data": self.left.__getstate__(),
            "right_data": right_data,
            "metadata": {
                "classname": self.__class__.__name__,
                "module_name": module_name,
                "module_version": version,
                "serialization_date": datetime.now().isoformat(),
                "python_version": sys.version,
            },
        }
        return state

    def __setstate__(self, state):
        """Set the current state of the covariance from deserialized state.

        This method is called when deserializing the predictor instance. It
        takes a dictionary that represents the state of the covariance function and
        updates the functions's attributes accordingly.

        :param state: A dictionary representing the state of the covariance function.
        :type state: dict
        """
        if not isinstance(state, dict) or state.get("type") != "mellon.Covariance":
            raise ValueError(
                "The passed dict does not seem to define a covariance kernel."
            )
        self.left = Covariance.from_dict(state["left_data"])
        if (
            isinstance(state["right_data"], dict)
            and state["right_data"].get("type") == "mellon.Covariance"
        ):
            self.right = Covariance.from_dict(state["right_data"])
        else:
            self.right = deserialize(state["right_data"])


class Add(CovariancePair):
    R"""
    Supports adding covariance functions with + operator.
    """

    def __repr__(self):
        return "(" + repr(self.left) + " + " + repr(self.right) + ")"

    def k(self, x, y):
        if callable(self.right):
            return self.left(x, y) + self.right(x, y)
        return self.left(x, y) + self.right


class Mul(CovariancePair):
    R"""
    Supports multiplying covariance functions with * operator.
    """

    def __repr__(self):
        return "(" + repr(self.left) + " * " + repr(self.right) + ")"

    def k(self, x, y):
        if callable(self.right):
            return self.left(x, y) * self.right(x, y)
        return self.left(x, y) * self.right


class Pow(CovariancePair):
    R"""
    Supports taking a covariance function to a power with ** operator.
    """

    def __repr__(self):
        return "(" + repr(self.left) + " ** " + repr(self.right) + ")"

    def k(self, x, y):
        return self.left(x, y) ** self.right



from .cov import (
    Matern32,
    Matern52,
    ExpQuad,
    Exponential,
    RatQuad,
)
from .base_cov import (
    Covariance,
    CovariancePair,
)

__all__ = [
    "Covariance",
    "CovariancePair",
    "Matern32",
    "Matern52",
    "ExpQuad",
    "Exponential",
    "RatQuad",
]



from setuptools import setup

# read the contents of your README file
from pathlib import Path

this_directory = Path(__file__).parent


def get_version(rel_path):
    for line in (this_directory / rel_path).read_text().splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setup(
    name="mellon",
    version=get_version("mellon/__init__.py"),
    description="Non-parametric density estimator.",
    url="https://github.com/settylab/mellon",
    author="Setty Lab",
    author_email="msetty@fredhutch.org",
    license="GNU General Public License v3.0",
    packages=["mellon"],
    install_requires=[
        "jax",
        "jaxopt",
        "scikit-learn",
    ],
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    long_description=(this_directory / "README.rst").read_text(),
    long_description_content_type="text/x-rst",
)



from jax.numpy import cumsum, searchsorted, count_nonzero, sqrt, isnan, any
from jax.numpy.linalg import eigh, cholesky, qr
from jax.scipy.linalg import solve_triangular
from .util import stabilize, DEFAULT_JITTER, Log


DEFAULT_RANK = 1.0
DEFAULT_METHOD = "auto"

logger = Log()


def _check_method(rank, full, method):
    R"""
    Checks if rank is a float 0.0 :math:`\le` rank :math:`\le` 1.0 or an int
    1 :math:`\le` rank :math:`\le` full. Raises an error if neither is true
    or if method doesn't match the detected method.

    :param rank: The rank of the decomposition, or if rank is a float greater
    than 0 and less than 1, the rank is reduced further using the QR decomposition
    such that the eigenvalues of the included eigenvectors account for the
    specified percentage of the total eigenvalues. Defaults to 0.999.
    :type rank: int or float
    :param full: The size of the exact matrix.
    :type full: int
    :param method: The method to interpret the rank.
    :type method: str
    :return: method - The detected method.
    :rtype: str
    """

    percent = isinstance(rank, float) and (0 < rank) and (rank <= 1)
    fixed = isinstance(rank, int) and (1 <= rank) and (rank <= full)
    if not (percent or fixed):
        message = """rank must be a float 0.0 <=rank <= 1.0 or
    an int 1 <= rank <= q. q equals the number of landmarks
    or the number of data points if there are no landmarks."""
        raise ValueError(message)
    elif percent and not (method == "percent" or method == "auto"):
        message = f"""The argument method={method} does not match the rank={rank}.
    The detected method from the rank is 'percent'."""
        raise ValueError(message)
    elif fixed and not (method == "fixed" or method == "auto"):
        message = f"""The argument method={method} does not match the rank={rank}.
    The detected method from the rank is 'fixed'."""
        raise ValueError(message)
    if percent:
        return "percent"
    else:
        return "fixed"


def _eigendecomposition(A, rank=DEFAULT_RANK, method=DEFAULT_METHOD):
    R"""
    Decompose :math:`A` into its largest positive `rank` and
    at least one eigenvector(s) and eigenvalue(s).

    :param A: A square matrix.
    :type A: array-like
    :param rank: The rank of the decomposition, or if rank is a float
        0.0 :math:`\le` rank :math:`\le` 1.0, the rank is reduced further using the QR
        decomposition such that the eigenvalues of the included eigenvectors account for
        the specified percentage of the total eigenvalues. Defaults to 0.999.
    :type rank: int or float
    :param method: Explicitly specifies whether rank is to be interpreted as a
        fixed number of eigenvectors or a percent of eigenvalues to include
        in the low rank approximation.
    :type method: str
    :return: :math:`s, v` - The top eigenvalues and eigenvectors.
    :rtype: array-like, array-like
    """

    s, v = eigh(A)
    if any(s <= 0):
        message = (
            "Singuarity detected in covariance matrix. "
            "This can complicated prediction. Consider raising the jitter."
        )
        logger.warning(message)
    p = count_nonzero(s > 0)  # stability
    summed = cumsum(s[: -p - 1 : -1])
    if method == "percent":
        # automatically choose rank to capture some percent of the eigenvalues
        target = summed[-1] * rank
        p = searchsorted(summed, target)
        if p == 0:
            logger.warning(
                f"Low variance percentage {rank:%} indicated rank=0. "
                "Bumping rank to 1."
            )
            p = 1
    else:
        p = min(rank, p)
    if (method == "percent" and rank < 1) or rank < len(summed):
        frac = summed[p] / summed[-1]
        logger.info(f"Recovering {frac:%} variance in eigendecomposition.")
    s_ = s[-p:]
    v_ = v[:, -p:]
    return s_, v_


def _full_rank(x, cov_func, jitter=DEFAULT_JITTER):
    R"""
    Compute :math:`L` such that :math:`L L^\top = K`, where :math:`K`
    is the full rank covariance matrix.

    :param x: The training instances.
    :type x: array-like
    :param cov_func: The Gaussian process covariance function.
    :type cov_func: function
    :param jitter: A small amount to add to the diagonal. Defaults to 1e-6.
    :type jitter: float
    :return: :math:`L` - A matrix such that :math:`L L^\top = K`.
    :rtype: array-like
    """
    W = stabilize(cov_func(x, x), jitter)
    L = cholesky(W)
    if any(isnan(L)):
        message = (
            f"Covariance not positively definite with jitter={jitter}. "
            "Consider increasing the jitter for numerical stabilization."
        )
        logger.error(message)
        raise ValueError(message)
    return L


def _full_decomposition_low_rank(
    x, cov_func, rank=DEFAULT_RANK, method=DEFAULT_METHOD, jitter=DEFAULT_JITTER
):
    R"""
    Compute a low rank :math:`L` such that :math:`L L^\top ~= K`, where :math:`K` is the
    full rank covariance matrix. The rank is less than or equal to the number of
    landmark points.

    :param x: The training instances.
    :type x: array-like
    :param cov_func: The Gaussian process covariance function.
    :type cov_func: function
    :param rank: The rank of the decomposition, or if rank is a float greater
        than 0 and less than 1, the eigenvalues of the included eigenvectors
        account for the specified percentage of the total eigenvalues.
        Defaults to 0.999.
    :type rank: int or float
    :param jitter: A small amount to add to the diagonal. Defaults to 1e-6.
    :type jitter: float
    :param method: Explicitly specifies whether rank is to be interpreted as a
        fixed number of eigenvectors or a percent of eigenvalues to include
        in the low rank approximation. Supports 'fixed', 'percent', or 'auto'.
        If 'auto', interprets rank as a fixed number of eigenvectors if it is
        an int and interprets rank as a percent of eigenvalues if it is a float.
        Defaults to 'auto'.
    :type method: str
    :return: :math:`L` - A matrix such that :math:`L L^\top \approx K`.
    :rtype: array-like
    """
    W = cov_func(x, x)
    s, v = _eigendecomposition(W, rank=rank, method=method)
    L = v * sqrt(s)
    return L


def _standard_low_rank(x, cov_func, xu, jitter=DEFAULT_JITTER):
    R"""
    Compute a low rank :math:`L` such that :math:`L L^\top \approx K`, where :math:`K`
    is the full rank covariance matrix. The rank is equal to the number of
    landmark points.

    :param x: The training instances.
    :type x: array-like
    :param cov_func: The Gaussian process covariance function.
    :type cov_func: function
    :param xu: The landmark points.
    :type xu: array-like
    :param jitter: A small amount to add to the diagonal. Defaults to 1e-6.
    :type jitter: float
    :return: :math:`L` - A matrix such that :math:`L L^\top \approx K`.
    :rtype: array-like
    """
    W = stabilize(cov_func(xu, xu), jitter)
    C = cov_func(x, xu)
    U = cholesky(W)
    if any(isnan(U)):
        message = (
            f"Covariance of landmarks not positively definite with jitter={jitter}. "
            "Consider increasing the jitter for numerical stabilization."
        )
        logger.error(message)
        raise ValueError(message)
    L = solve_triangular(U, C.T, lower=True).T
    return L


def _modified_low_rank(
    x, cov_func, xu, rank=DEFAULT_RANK, method=DEFAULT_METHOD, jitter=DEFAULT_JITTER
):
    R"""
    Compute a low rank :math:`L` such that :math:`L L^\top ~= K`, where :math:`K` is the
    full rank covariance matrix. The rank is less than or equal to the number of
    landmark points.

    :param x: The training instances.
    :type x: array-like
    :param cov_func: The Gaussian process covariance function.
    :type cov_func: function
    :param xu: The landmark points.
    :type xu: array-like
    :param rank: The rank of the decomposition, or if rank is a float
        0.0 :math:`\le` rank :math:`\le` 1.0, the rank is reduced further using
        the QR decomposition such that the eigenvalues of the included eigenvectors
        account for the specified percentage of the total eigenvalues. Defaults to 0.999.
    :type rank: int or float
    :param jitter: A small amount to add to the diagonal. Defaults to 1e-6.
    :type jitter: float
    :param method: Explicitly specifies whether rank is to be interpreted as a
        fixed number of eigenvectors or a percent of eigenvalues to include
        in the low rank approximation. Supports 'fixed', 'percent', or 'auto'.
        If 'auto', interprets rank as a fixed number of eigenvectors if it is
        an int and interprets rank as a percent of eigenvalues if it is a float.
        Defaults to 'auto'.
    :type method: str
    :return: :math:`L` - A matrix such that :math:`L L^\top \approx K`.
    :rtype: array-like
    """
    W = stabilize(cov_func(xu, xu), jitter)
    C = cov_func(x, xu)
    Q, R = qr(C, mode="reduced")
    s, v = _eigendecomposition(W, rank=xu.shape[0], method="fixed")
    T = R @ v
    S, V = _eigendecomposition(T / s @ T.T, rank=rank, method=method)
    L = Q @ V * sqrt(S)
    return L



from .decomposition import DEFAULT_METHOD
from .base_model import BaseEstimator, DEFAULT_COV_FUNC
from .inference import (
    compute_conditional_mean,
    DEFAULT_N_ITER,
    DEFAULT_INIT_LEARN_RATE,
    DEFAULT_OPTIMIZER,
)
from .parameters import (
    DEFAULT_N_LANDMARKS,
)
from .util import (
    DEFAULT_JITTER,
    Log,
)
from .validation import (
    _validate_positive_float,
    _validate_float,
    _validate_array,
)


DEFAULT_D_METHOD = "embedding"

logger = Log()


class FunctionEstimator(BaseEstimator):
    R"""
    This class implements a Function Estimator that uses a conditional normal distribution
    to smoothen and extend a function on all cell states using the Mellon abstractions.

    Parameters
    ----------
    cov_func_curry : function or type
        A curry that takes one length scale argument and returns a covariance function
        of the form k(x, y) :math:`\rightarrow` float. Defaults to Matern52.

    n_landmarks : int, optional
        The number of landmark points. If less than 1 or greater than or equal to the
        number of training points, inducing points will not be computed or used. Defaults to 5000.

    jitter : float, optional
        A small amount added to the diagonal of the covariance matrix to ensure numerical stability.
        Defaults to 1e-6.

    landmarks : array-like or None, optional
        Points used to quantize the data for the approximate covariance. If None, landmarks are
        set as k-means centroids with k=n_landmarks. This is ignored if n_landmarks is greater than
        or equal to the number of training points. Defaults to None.

    nn_distances : array-like or None, optional
        The nearest neighbor distances at each data point. If None, computes the nearest neighbor
        distances automatically, with a KDTree if the dimensionality of the data is less than 20,
        or a BallTree otherwise. Defaults to None.

    mu : float
        The mean of the Gaussian process :math:`\mu`. Defaults to 0.

    ls : float or None, optional
        The length scale for the Gaussian process covariance function.
        If None (default), the length scale is automatically selected based on
        a heuristic link between the nearest neighbor distances and the optimal
        length scale.

    ls_factor : float, optional
        A scaling factor applied to the length scale when it's automatically
        selected. It is used to manually adjust the automatically chosen length
        scale for finer control over the model's sensitivity to variations in the data.

    cov_func : function or None, optional
        The Gaussian process covariance function of the form k(x, y) :math:`\rightarrow` float.
        If None, automatically generates the covariance function cov_func = cov_func_curry(ls).
        Defaults to None.

    sigma : float, optional
        The standard deviation of the white noise. Defaults to 0.

    jit : bool, optional
        Use JAX just-in-time compilation for the loss function and its gradient during optimization.
        Defaults to False.
    """

    def __init__(
        self,
        cov_func_curry=DEFAULT_COV_FUNC,
        n_landmarks=DEFAULT_N_LANDMARKS,
        method=DEFAULT_METHOD,
        jitter=DEFAULT_JITTER,
        optimizer=DEFAULT_OPTIMIZER,
        n_iter=DEFAULT_N_ITER,
        init_learn_rate=DEFAULT_INIT_LEARN_RATE,
        landmarks=None,
        nn_distances=None,
        mu=0,
        ls=None,
        ls_factor=1,
        cov_func=None,
        sigma=0,
        jit=True,
    ):
        super().__init__(
            cov_func_curry=cov_func_curry,
            n_landmarks=n_landmarks,
            rank=1.0,
            jitter=jitter,
            landmarks=landmarks,
            nn_distances=nn_distances,
            mu=mu,
            ls=ls,
            ls_factor=ls_factor,
            cov_func=cov_func,
            jit=jit,
        )
        self.mu = _validate_float(mu, "mu")
        self.sigma = _validate_positive_float(sigma, "sigma")

    def __call__(self, x=None, y=None):
        """This calls self.fit_predict(x, y):
        Compute the conditional mean and return the smoothed function values
        at the points x.

        :param x: The training instances to estimate function.
        :type x: array-like
        :param y: The training function values on cell states.
        :type y: array-like
        :return: condition_mean - The conditional mean function value at each test point in x.
        :rtype: array-like
        """
        return self.fit_predict(x=x, y=y)

    def prepare_inference(self, x):
        R"""
        Set all attributes in preparation. It is not necessary to call this
        function before calling fit.

        :param x: The cell states.
        :type x: array-like
        :param y: The function values on the cell states.
        :type y: array-like
        :return: loss_func, initial_value - The Bayesian loss function and
            initial guess for optimization.
        :rtype: function, array-like
        """
        x = self.set_x(x)
        if self.ls is None:
            self._prepare_attribute("nn_distances")
        self._prepare_attribute("ls")
        self._prepare_attribute("cov_func")
        self._prepare_attribute("landmarks")
        return

    def compute_conditional(self, x=None, y=None):
        R"""
        Compute and return the conditional mean function.

        :param x: The training instances to estimate density function.
        :type x: array-like
        :param y: The training function values on cell states.
        :type y: array-like
        :return: condition_mean_function - The conditional mean function.
        :rtype: array-like
        """
        if x is None:
            x = self.x
        else:
            x = _validate_array(x, "x")
        if self.x is not None and self.x is not x:
            message = (
                "self.x has been set already, but is not equal to the argument x. "
                "Current landmarks might be inapropriate."
            )
            logger.warning(message)
        if self.x is None and x is None:
            message = "Required argument x is missing and self.x has not been set."
            raise ValueError(message)
        if y is None:
            message = "Required argument y is missing."
            raise ValueError(message)
        landmarks = self.landmarks
        mu = self.mu
        cov_func = self.cov_func
        sigma = self.sigma
        jitter = self.jitter
        conditional = compute_conditional_mean(
            x,
            landmarks,
            None,
            y,
            mu,
            cov_func,
            sigma,
            jitter=jitter,
        )
        self.conditional = conditional
        return conditional

    def fit(self, x=None, y=None):
        """
        Trains the model using the provided training data and function values. This includes preparing
        the model for inference and computing the conditional distribution for the given data.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_features), default=None
            The training instances where `n_samples` is the number of samples and `n_features` is the
            number of features. Each sample is an array of features representing a point in the feature space.

        y : array-like of shape (n_samples, n_output_features), default=None
            The function values of the training instances. `n_samples` is the number of samples and
            `n_output_features` is the number of function values at each sample.

        Raises
        ------
        ValueError
            If the number of samples in `x` and `y` doesn't match.

        Returns
        -------
        self : object
            This method returns self for chaining.
        """
        x = self.set_x(x)
        y = _validate_array(y, "y")

        n_samples = x.shape[0]
        # Check if the number of samples in x and y match
        if y.shape[0] != n_samples:
            raise ValueError(
                f"X.shape[0] = {n_samples} (n_samples) should equal "
                "y.shape[0] = {y.shape[0]}."
            )

        self.prepare_inference(x)
        self.compute_conditional(x, y)
        return self

    @property
    def predict(self):
        """
        A property that returns an instance of the :class:`mellon.Predictor` class. This predictor can be used
        to predict the function values for new data points by calling the instance like a function.

        The predictor instance also supports serialization features, allowing for saving and loading the
        predictor's state. For more details, refer to the :class:`mellon.Predictor` documentation.

        Returns
        -------
        mellon.Predictor
            A predictor instance that computes the conditional mean function value at each new data point.

        Example
        -------

        >>> y_pred = model.predict(Xnew)

        """
        return self.conditional

    def fit_predict(self, x=None, y=None, Xnew=None):
        """
        Trains the model using the provided training data and function values, then makes predictions
        for new data points. The function computes the conditional mean and returns the smoothed function
        values at the points `Xnew` for each column of values in `y`.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_features), default=None
            The training instances where `n_samples` is the number of samples and `n_features` is the number
            of features. Each sample is an array of features representing a point in the feature space.

        y : array-like of shape (n_samples, n_output_features), default=None
            The function values of the training instances. `n_samples` is the number of samples and
            `n_output_features` is the number of function values at each sample.

        Xnew : array-like of shape (n_predict_samples, n_features), default=None
            The new data points to make predictions on where `n_predict_samples` is the number of samples
            to predict and `n_features` is the number of features. If not provided, the predictions will
            be made on the training instances `x`.

        Returns
        -------
        array-like of shape (n_predict_samples, n_output_features)
            The conditional mean function value at each new data point in `Xnew`. The number of predicted
            function values at each sample will match the number of output features in `y`.

        Raises
        ------
        ValueError
            If the number of samples in `x` and `y` don't match, or if the number of features in `x` and
            `Xnew` don't match.
        """
        x = self.set_x(x)
        y = _validate_array(y, "y")
        Xnew = _validate_array(Xnew, "Xnew", optional=True)

        # If Xnew is not provided, default to x
        if Xnew is None:
            Xnew = x
        else:
            # Ensure the number of dimensions in x and Xnew are the same
            if x.ndim != Xnew.ndim:
                raise ValueError(
                    f"The provided arrays, 'x' and 'Xnew', do not have the same number of dimensions. "
                    f"'x' is {x.ndim}-D and 'Xnew' is {Xnew.ndim}-D. Please provide arrays with consistent dimensionality."
                )

            # If both arrays are multi-dimensional, ensure they have the same number of features
            if x.ndim > 1 and x.shape[1] != Xnew.shape[1]:
                raise ValueError(
                    f"The provided arrays, 'x' and 'Xnew', should have the same number of features. "
                    f"Got Xnew.shape[1] = {Xnew.shape[1]}, but expected it to be equal to x.shape[1] = {x.shape[1]}. "
                    "Please provide arrays with the same number of features."
                )

        # Fit the model and predict
        self.fit(x, y)
        return self.predict(Xnew)

    def multi_fit_predict(self, x=None, Y=None, Xnew=None):
        """
        Compute the conditional mean and return the smoothed function values
        at the points Xnew for each line of values in Y.

        This method is deprecated. Use FunctionEstimator.fit_reodict instead.

        Parameters
        ----------
        x : array-like, optional
            The training instances to estimate density function.
        Y : array-like, optional
            The training function values on cell states.
        Xnew : array-like, optional
            The new data to predict.

        Returns
        -------
        array-like
            The conditional mean function value at each test point in `x`.

        Raises
        ------
        ValueError
            If the number of samples in `x` and `Y` don't match, or if the
            number of features in `x` and `Xnew` don't match.
        """

        logger.warning(
            "Deprecation Warning: FunctionEstimator's multi_fit_predict method is deprecated. "
            "Use FunctionEstimator.fit_reodict instead."
        )

        # Set the x and validate inputs
        x = self.set_x(x)
        Y = _validate_array(Y, "Y")

        n_samples = x.shape[0]

        # Check for consistency in sample size between x and Y
        if Y.shape[0] != n_samples:
            if Y.shape[1] == n_samples:
                logger.warning(
                    "Y.shape[0] does not equal X.shape[0] (the number of samples). "
                    "However, Y.shape[1] == X.shape[0]. Transposing Y. "
                    "This assumes the columns of Y are the samples. Please verify."
                )
                Y = Y.T
        return self.fit_predict(x, Y, Xnew).T
